
function PricesEGM(K,PHI,τ,Z,params::AiyagariParametersEGM) # K is here K(-1)
#before tax
    @unpack β,α,δ,γ,χ,ϕ,Se = params
    CC = Se*(χ*(1-τ))^ϕ
    w = (Z*(1-α)*(K/CC)^α)^(1/(1+ϕ*α))
    L = CC*(w^ϕ)
    R = α*Z*(K/L)^(α-1.0) + 1.0 - δ
      return AggVarsEGM(R,w,PHI,τ)
end


function AiyagariEGM(
    K::T,
    ns::I,
    Transv::Array{T,1},
    Transe::Array{T,1},
    states::Array{T,1}, 
    amin::T = 1e-9,
    amax::T = 200.0,
    curv::Real = 4,
    ϕ::T = 0.1,  
    χ::T = 330.0, 
    β::T = 0.98,
    α::T = 0.4,
    δ::T = 0.02,
    γ::T = 1.5,    
    PHI::T = 0.0,
    τ::T = 0.0,     
    Se::T = 0.95,
    na::I = 100,
    nd::I = 100) where{T <: Real,I <: Integer}

    #############Params
    params = AiyagariParametersEGM(β,α,δ,γ,Se,χ,ϕ)
    AggVars = PricesEGM(K,PHI,τ,1.0,params)
    @unpack R,w = AggVars

    ################## Policy grid
    function grid_fun(a_min,a_max,na, pexp)
        x = range(a_min,step=0.5,length=na)
        grid = a_min .+ (a_max-a_min)*(x.^pexp/maximum(x.^pexp))
        return grid
    end
    aGrid = grid_fun(amin,amax,na,curv)
    ################### Distribution grid
    #dGrid = collect(range(aGrid[1],stop = aGrid[end],length = nd))
    #aGrid = collect(range(amin,stop = amax,length = na))
    dGrid=aGrid

   
  #  guess = vcat(10.0 .+ aGrid,10.0 .+ aGrid)
 #   guess = kron(ones(Float64,ns),10 .+ aGrid)
  #  aGridl_t = kron(ones(Float64,ns),aGrid)
  guess = repeat(10 .+ aGrid,ns*2)
  aGridl_t = repeat(aGrid,ns*2)    
    

    #return AiyagariModelEGM(params,aGrid,vcat(aGrid,aGrid),na,dGrid,nd,states,ns,Transv),guess,AggVars
    return AiyagariModelEGM(params,aGrid,aGridl_t,na,dGrid,nd,states,ns,Transv,Transe),guess,AggVars
end




function interpEGM(pol::AbstractArray,
                grid::AbstractArray,
                x::T,
                na::Integer) where{T <: Real}
    np = searchsortedlast(pol,x)

    # if you give x (current asset) (out of the policy pol), gives ap (saving decision)  out of the grid (grid)
    
    ##Adjust indices if assets fall out of bounds
    (np > 0 && np < na) ? np = np : 
        (np == na) ? np = na-1 : 
            np = 1        
    #@show np
    ap_l,ap_h = pol[np],pol[np+1]
    a_l,a_h = grid[np], grid[np+1] 
    ap = a_l + (a_h-a_l)/(ap_h-ap_l)*(x-ap_l) 
    
    above =  ap > 0.0 
    return above*ap,np
    # return either 0 or ap
end


function get_ZEGM(pol::AbstractArray,
               Aggs::AggVarsEGM,
               CurrentAssets::AbstractArray,
               AiyagariModel::AiyagariModelEGM,
               cpol::AbstractArray)
                
    @unpack params,aGrid,na,ns,states = AiyagariModel
    @unpack  χ, ϕ = params
    
    τ = Aggs.τ
    PHI = Aggs.PHI
    
    #X = copy(states)
    Xe = ((χ*(1-τ)*Aggs.w .*states).^(1+ ϕ))/χ 
    # Xu = ((χ*Aggs.PHI*Aggs.w .*states).^(1+ ϕ))/χ 
    #Xc = (1 -ϕ/(1+ϕ)).*[Xu;Xe] #corrected
    Xc =[ (PHI/(1-τ) -ϕ/(1+ϕ)).*Xe;(1 -ϕ/(1+ϕ)).*Xe] #corrected
    Z = copy(cpol)
    pol = reshape(pol,na,ns,2) # to interpolate on asset only
    for si = 1:ns
        for ai = 1:na
            for k = 1:2 # 1 is unemployed get PHI times...
            asi = (si - 1)*na + ai + (k-1)*ns*na
            
            Z[asi] = Aggs.R*CurrentAssets[asi] + Xc[si+(k-1)*ns] - interpEGM(pol[:,si,k],aGrid,CurrentAssets[asi],na)[1] # compact form
            end
       end
    end
    return Z
end

#        pol = EulerBackEGM(pol,Pr,Pr_P,AiyagariModel,cmat,cmat2)[1]

function EulerBackEGM(pol::AbstractArray,
                   Aggs::AggVarsEGM,
                   Aggs_P::AggVarsEGM, # for transitions
                   AiyagariModel::AiyagariModelEGM,
                   cpol::AbstractArray,
                   apol::AbstractArray)
    
    @unpack params,na,nd,ns,aGridl,Transv,Transe,states = AiyagariModel
    @unpack γ,β,χ,ϕ = params

    Trans = reshape(Transv,ns,ns)
    Transee = reshape(Transe,2,2)
  
    
    R_P,w_P,PHI_P,τ_P = Aggs_P.R,Aggs_P.w,Aggs_P.PHI,Aggs_P.τ
    R,w,PHI,τ = Aggs.R,Aggs.w,Aggs.PHI,Aggs.τ
    
   # cp = get_cEGM(pol,Aggs_P,aGridl,AiyagariModel,cpol) #vector ns*na
    Zp  = get_ZEGM(pol,Aggs_P,aGridl,AiyagariModel,cpol) #vector ns*na*2, future prices
    
    Zp = max.(Zp,0.001)   # may be negative in the iteration
    

    upcp = uPrime(Zp,γ)
    Eupcp = copy(cpol)
    #Eupcp_sp = 0.0
   
    #X = copy(states)
   # X = ((χ*(1-τ)*Aggs.w .*states).^(1+ ϕ))/χ
    Xe = ((χ*(1-τ)*Aggs.w .*states).^(1+ ϕ))/χ 
    Xu = ((χ*Aggs_P.PHI*Aggs.w .*states).^(1+ ϕ))/χ 
    X = [Xu;Xe]
    
    Lse = (χ*(1-τ)*Aggs.w .*states).^ϕ   # labor supply by type (employed)
    Lsu = (χ*PHI*Aggs.w .*states).^ϕ     # labor supply by type (unemployed)    for utility
    Ls = [Lse;Lse]
    for ai = 1:na
        for si = 1:ns
            for k=1:2
           
                asi = (si-1)*na + ai + (k-1)*na*ns
            Eupcp_sp = 0.0
            for spi = 1:ns
                    for kp = 1:2
                    aspi = (spi-1)*na + ai + (kp-1)*na*ns
                    Eupcp_sp += Transee[kp,k]*Trans[spi,si]*upcp[aspi]
                    end
                    
            end
            Eupcp[asi] = Eupcp_sp 
            
            end
        end
    end

    upc = β*Eupcp*R_P #OK next period price
    
    c = copy(upc) # to define the dimension
    Z = uPrimeInv(upc,γ)

    Xe = ((χ*(1-τ)*Aggs.w .*states).^(1+ ϕ))/χ #current prices
    X = [Xe;Xe]
    Xc = [(PHI/(1-τ))*Xe;Xe]
    for ai = 1:na
        for si = 1:ns
           for k=1:2
            asi       = (si-1)*na + ai + (k-1)*na*ns
            c[asi]    = Z[asi] + X[si+(k-1)*ns]*ϕ/(1+ϕ)
            apol[asi] = (aGridl[asi] + c[asi] - Xc[si+(k-1)*ns])/R                  
            
            end
        end
    end
     UU = Uvv.(Z)
    
    return apol,c,upc,Ls,UU
end


function SolveEGM(pol::AbstractArray,
                  Aggs::AggVarsEGM,
                  AiyagariModel::AiyagariModelEGM,
                  cpol::AbstractArray,
                  apol::AbstractArray,tol = 1e-17)
    @unpack ns,na = AiyagariModel

    for i = 1:10000
        a = EulerBackEGM(pol,Aggs,Aggs,AiyagariModel,cpol,apol)[1]
        if (i-1) % 50 == 0
            test = abs.(a - pol)/(abs.(a) + abs.(pol))
           # println("iteration: ",i," ",maximum(test))
            if maximum(test) < tol
                println("Solved in ",i," ","iterations")
                break
            end
        end
        pol = copy(a)
    end
    return pol
end

#########################################################
 # Below the functions  for the stationary distribution
#########################################################

function MakeTransMatEGM(pol,AiyagariModel,tmat)
    @unpack ns,na,aGrid,Transv,Transe = AiyagariModel
    pol = reshape(pol,na,ns*2)
    
    Trans   = reshape(Transv,ns,ns)
    Transee = reshape(Transe,2,2)
    tmat = copy(tmat)
    tmat  .=0.0

    for a_i = 1:na
        for j = 1:ns
            for e_i=1:2 #employment status
                x,i = interpEGM(pol[:,j+(e_i-1)*ns],aGrid,aGrid[a_i],na)
                p = (aGrid[a_i] - pol[i,j+(e_i-1)*ns])/(pol[i+1,j+(e_i-1)*ns] - pol[i,j+(e_i-1)*ns])
                p = min(max(p,0.0),1.0)
                sj = a_i+(j-1)*na + (e_i-1)*na*ns
                for k = 1:ns
                    for e_p=1:2
                        sk = i+(k-1)*na + (e_p-1)*na*ns
                        tmat[sk+1,sj] = p * Trans[k,j] * Transee[e_p,e_i]
                        tmat[sk,sj] = (1.0-p) * Trans[k,j] * Transee[e_p,e_i]
                    end
                end
            end
    
        end     
    end
   
    return tmat
end

function StationaryDistributionEGM(Mat,AiyagariModel)
    @unpack params,ns,nd,na = AiyagariModel 
    x = zeros(Float64,(na*ns*2,1))
    λ, x = powm!(Mat, ones(ns*na*2)/(ns*na*2), maxiter = 100000,tol = 1e-8)
    x = x/sum(x)
    return x
end


function steady(K0::T,
    initialpol::AbstractArray,
    AiyagariModel::AiyagariModelEGM,
    R::T,
    w::T,    
    PHI::T = 0.0,    
    τ::T = 0.0,
    tol = 1e-10,maxn = 50)where{T <: Real}
    @unpack params,aGrid,na,dGrid,nd,ns,na = AiyagariModel
    
    tmat = zeros(eltype(initialpol),(na*ns*2,na*ns*2))
    cmat = zeros(eltype(initialpol),na*ns*2)
    pol = initialpol
    
   # Aggs = AggVarsEGM(R,w,PHI,τ)
   # cmat .= 0.0
   # pol = SolveEGM(pol,Aggs,AiyagariModel,cmat,cmat)

    #Stationary transition
 
    
    
    #Aggregate savings
   # dGridl_t = repeat(dGrid,ns*2)
  #  EA = dot(dGridl_t,D)
   
    #println(" Int. rate: ",Aggs.R," Bond S.: ",EA)
    
   
     relax  = 5e-2
     uK,lK = 2*K0, 0.0
    
    
    
    
    pol = initialpol
    
    
    for kit = 1:maxn
        Aggs = PricesEGM(K0,PHI,τ,1.0,params) ##steady state
        
       #  Aggs = AggVarsEGM(R,w,PHI,τ)
         cmat .= 0.0
         pol = SolveEGM(pol,Aggs,AiyagariModel,cmat,cmat)

        
        
        
        #Stationary transition
        tmat .= 0.0
        trans = MakeTransMatEGM(pol,AiyagariModel,tmat)
        D = StationaryDistributionEGM(trans,AiyagariModel)

        #Aggregate savings
        
        #EA = dot(vcat(dGrid,dGrid),D)
        #dGridl_t = kron(ones(Float64,ns),dGrid)
        dGridl_t = repeat(dGrid,ns*2)
        EA = dot(dGridl_t,D)
        
        
        if (EA > K0) ### too little lending -> low r -> too much borrowing 
            uK = min(EA,uK)  
            lK = max(K0,lK)
            K0 = 1.0/2.0*(lK + uK)
        else ## too much lending -> high r -> too little borrowing
            uK = min(K0,uK)
            lK = max(EA,lK)
            K0 = 1.0/2.0*(lK + uK)
        end
 
        # Only start convergence
    #   K0 = (1 - relax)*K0 + relax*EA
        
        
        println("It. ",kit," Int. rate: ",Aggs.R," Bond S.: ",EA, " Bond D.: ",K0)
       # @show K0
        if abs(EA - K0) < tol
            println("Markets clear!")
            #println("Interest rate: ",R," ","Bonds: ",EA)
            cmat .= 0.0
            polA,polC,upc,Ls,Ulevel = EulerBackEGM(pol,Aggs,Aggs,AiyagariModel,cmat,cmat)
            return polA,polC,D,EA,Aggs,trans,upc,Ls,Ulevel  
            break
          
        end
    end
    
    return println("Markets did not clear")
    
    
   # cmat .= 0.0
   # polA,polC,upc = EulerBackEGM(pol,Aggs,Aggs,AiyagariModel,cmat,cmat)
    
    
    
    
   # return polA,polC,D,EA,Aggs,trans,upc  
end









function Euler_err(       
        N::Integer, # length of the truncation
        AA0::AiyagariModelEGM,
        polA_ss::Array{T,1},
        polC_ss::Array{T,1},
        upc::Array{T,1},
        D_ss::Array{T,1},         
        Aggs_ss::AggVarsEGM{T,T},
        Mat::Array{T,2},
        A::T,
        vstan::Array{T,1},
        Ls::Array{T,1},
        Ulevel::Array{T,1},
        states::Array{T,1}) where {T <: Real}

#polA_ss,polC_ss,D_ss,A,Aggs_ss,Mat,upc = steady(K0,pol0,AA0,R,w,PHI,τ,1e-4);    
    @unpack params,aGrid,na,dGrid,nd,ns,Transv,Transe,states = AA0
    @unpack γ,β,χ,ϕ = params
    
    R = Aggs_ss.R
    w = Aggs_ss.w
    τ = Aggs_ss.τ
    Se = sum(vstan[ns+1:end]) # share of employed worker
    Ltot = Se*(χ*(1-τ)*w)^ϕ
    
#################    #################    #################    #################     
    
Va = repeat(aGrid,ns*2)'*Mat # policy rule in a a'(a)
pol = reshape(Va,na,ns*2)

γ =  AA0.params.γ
β =  AA0.params.β

Trans   = reshape(AA0.Transv,ns,ns)
Transee = reshape(AA0.Transe,2,2)


############################
# computing CC bins     
############################    
Vmat = Mat[1:na,:]
aGrid = AA0.aGrid
aGrid[1] = 0 
Va = repeat(aGrid,ns*2)'*Mat # policy rule in a a'(a)
pol = reshape(Va,na,ns*2)
Vind = zeros(Int,na*ns*2)
Wind = zeros(Float64,na*ns*2)
for i = 1:length(Va)
    indcc=findall(x->x<=Va[i],aGrid) 
    Vind[i] = maximum(indcc)
    Wind[i] = 1-(Va[i] - aGrid[Vind[i]])/(aGrid[Vind[i]+1]-aGrid[Vind[i]])
    
end
    
γ =  AA0.params.γ
β =  AA0.params.β
R = Aggs_ss.R
Trans   = reshape(AA0.Transv,ns,ns)
Transee = reshape(AA0.Transe,2,2)

Lv = [Aggs_ss.PHI*kron(Ls[1:ns],ones(na,1));(1-Aggs_ss.τ)*kron(Ls[1:ns],ones(na,1))]
ytype = repeat(kron(collect(StepRange(1, Int8(1), ns)),ones(Int,na)),2)
Cpol =  - Va' + Aggs_ss.R*repeat(aGrid,ns*2) + Aggs_ss.w*states[ytype].*Lv    


etype = [ones(Int,na*ns);2*ones(Int,na*ns)]
ytype = repeat(kron(collect(StepRange(1, Int8(1), ns)),ones(Int,na)),2)

Xe = ((χ*(1-τ)*Aggs_ss.w .*states).^(1+ ϕ))/χ 
Xc = (ϕ/(1+ϕ))*[Xe;Xe]    
Xc = kron(Xc,ones(na,1))  
    
   

    
resi = β*R*(Mat')*upc    
cimp = (resi).^(-1/γ) .+Xc 
     
    

### I construct the vector of 0 for constrained bin and 1 otherwise, using the transition matrix
gridv = repeat([1;zeros(na-1)],ns*2) # grid of 1 only if minimum wealt (ie. CC)
matt = kron(gridv,ones(1,na*ns*2))
CCind = findall(x->x<=amin,(gridv.*Mat*ones(na*ns*2)))
Vc = 1.0 .-sum(matt.*Mat,dims=1)'

###  tector of residuals
EulerRes = Vc .*(Cpol./cimp .- 1.)     # not population weighted
    
    
    return EulerRes
end





